from __future__ import annotations

import json
import os
import shlex
import subprocess
import sys
import tempfile
import time
import webbrowser
from pathlib import Path
from typing import Any

import httpx
from InquirerPy import inquirer


DEFAULT_SERVER_URL = "https://alsoml.com"
CLAUDE_CONFIG_PATH = Path.home() / ".claude.json"
WORKOS_CLIENT_ID = "client_01KHFBZZ3JQF0NT8NVJ07GG932"


def _read_claude_config() -> dict[str, Any]:
    if not CLAUDE_CONFIG_PATH.exists():
        return {}
    try:
        return json.loads(CLAUDE_CONFIG_PATH.read_text())
    except (json.JSONDecodeError, OSError):
        return {}


def _write_claude_config(config: dict[str, Any]) -> None:
    fd, tmp_path = tempfile.mkstemp(
        dir=CLAUDE_CONFIG_PATH.parent, suffix=".tmp"
    )
    try:
        with os.fdopen(fd, "w") as f:
            json.dump(config, f, indent=2)
            f.flush()
            os.fsync(f.fileno())
        os.replace(tmp_path, str(CLAUDE_CONFIG_PATH))
    except BaseException:
        try:
            os.unlink(tmp_path)
        except OSError:
            pass
        raise


def _get_api_key() -> str:
    config = _read_claude_config()
    entry = config.get("mcpServers", {}).get("also-mcp", {})
    return entry.get("headers", {}).get("X-Api-Key", "")


def _get_server_url() -> str:
    config = _read_claude_config()
    entry = config.get("mcpServers", {}).get("also-mcp", {})
    url = entry.get("url", "")
    if url.endswith("/mcp"):
        return url[:-4]
    return DEFAULT_SERVER_URL


def _build_mcp_entry(server_url: str, api_key: str) -> dict[str, Any]:
    return {
        "type": "http",
        "url": f"{server_url}/mcp",
        "headers": {"X-Api-Key": api_key},
    }


def _sync_mcp_config(server_url: str, api_key: str) -> bool:
    if not api_key:
        return _remove_mcp_config()

    config = _read_claude_config()
    mcp_servers = config.setdefault("mcpServers", {})
    entry = _build_mcp_entry(server_url, api_key)
    mcp_servers["also-mcp"] = entry
    _write_claude_config(config)
    print(f"  Wrote MCP config to {CLAUDE_CONFIG_PATH}")
    print(f"  Entry: {json.dumps(entry, indent=2)}")
    return True


def _remove_mcp_config() -> bool:
    if not CLAUDE_CONFIG_PATH.exists():
        return False

    config = _read_claude_config()
    mcp_servers = config.get("mcpServers", {})
    if "also-mcp" not in mcp_servers:
        return False

    del mcp_servers["also-mcp"]
    if not mcp_servers:
        del config["mcpServers"]

    _write_claude_config(config)
    return True


def _api_headers(api_key: str) -> dict[str, str]:
    return {"X-Api-Key": api_key}


def _has_claude_descendant(pid: str) -> bool:
    try:
        ps_result = subprocess.run(
            ["ps", "-eo", "pid,ppid,command"],
            capture_output=True,
            text=True,
        )
    except FileNotFoundError:
        return False

    children_of: dict[str, list[tuple[str, str]]] = {}
    for line in ps_result.stdout.strip().splitlines()[1:]:
        parts = line.strip().split(None, 2)
        if len(parts) < 3:
            continue
        child_pid, parent_pid, cmd = parts
        children_of.setdefault(parent_pid, []).append((child_pid, cmd))

    stack = [pid]
    while stack:
        current = stack.pop()
        for child_pid, cmd in children_of.get(current, []):
            if "claude" in cmd.lower():
                return True
            stack.append(child_pid)
    return False


def _discover_claude_sessions() -> list[dict[str, str]]:
    result = subprocess.run(
        [
            "tmux", "list-panes", "-a",
            "-F", "#{session_name}:#{window_index}.#{pane_index} #{pane_pid} #{pane_current_command}",
        ],
        capture_output=True,
        text=True,
    )
    if result.returncode != 0:
        return []

    sessions: list[dict[str, str]] = []
    for line in result.stdout.strip().splitlines():
        parts = line.split(" ", 2)
        if len(parts) < 3:
            continue
        pane_id, pid, command = parts
        if "claude" not in command.lower() and not _has_claude_descendant(pid):
            continue
        session_name, rest = pane_id.split(":", 1)
        sessions.append({
            "pane_id": pane_id,
            "session_name": session_name,
            "pid": pid,
            "command": command,
        })
    return sessions


def _launch_new_claude_session() -> str | None:
    directory = inquirer.filepath(
        message="Directory to start Claude Code in:",
        default=str(Path.cwd()),
        only_directories=True,
    ).execute()
    if not directory:
        return None

    directory = str(Path(directory).resolve())

    has_tmux = subprocess.run(
        ["tmux", "list-sessions"],
        capture_output=True,
    ).returncode == 0

    if has_tmux:
        result = subprocess.run(
            [
                "tmux", "new-window",
                "-P", "-F", "#{session_name}:#{window_index}.#{pane_index}",
            ],
            capture_output=True,
            text=True,
        )
    else:
        result = subprocess.run(
            [
                "tmux", "new-session", "-d", "-s", "also",
                "-P", "-F", "#{session_name}:#{window_index}.#{pane_index}",
            ],
            capture_output=True,
            text=True,
        )

    if result.returncode != 0:
        print(f"  Failed to create tmux session: {result.stderr.strip()}", file=sys.stderr)
        return None

    pane_id = result.stdout.strip()
    subprocess.run(
        ["tmux", "send-keys", "-t", pane_id, f"cd {shlex.quote(directory)} && claude", "Enter"],
        check=True,
    )
    if not has_tmux:
        print("  Created tmux session 'also'. Run 'tmux attach -t also' to view it.")
    print(f"  Launched Claude Code in {directory} (pane {pane_id})")
    return pane_id


def _fetch_groups(server_url: str) -> list[str]:
    try:
        resp = httpx.get(f"{server_url}/auth/groups", timeout=10)
        resp.raise_for_status()
        return resp.json()
    except (httpx.HTTPError, json.JSONDecodeError):
        return []


def _create_group(server_url: str, group_id: str) -> str | None:
    try:
        resp = httpx.post(
            f"{server_url}/auth/groups",
            json={"group_id": group_id},
            timeout=10,
        )
        if resp.status_code == 409:
            print(f"  Group '{group_id}' already exists.", file=sys.stderr)
            return None
        resp.raise_for_status()
        return resp.json()["password"]
    except httpx.HTTPError as e:
        print(f"  Failed to create group: {e}", file=sys.stderr)
        return None


def _get_version() -> str:
    from importlib.metadata import version
    return version("alsogc")


def _authenticate_workos(server_url: str) -> tuple[str, str]:
    resp = httpx.post(
        "https://api.workos.com/user_management/authorize/device",
        data={"client_id": WORKOS_CLIENT_ID},
        timeout=30,
    )
    resp.raise_for_status()
    device_data = resp.json()

    device_code = device_data["device_code"]
    user_code = device_data["user_code"]
    verification_uri_complete = device_data["verification_uri_complete"]
    interval = device_data["interval"]

    print(f"\n  To sign in, visit: {verification_uri_complete}")
    print(f"  Enter code: {user_code}\n")

    try:
        webbrowser.open(verification_uri_complete)
    except Exception:
        pass

    while True:
        time.sleep(interval)
        print(f"  Polling... (interval={interval}s)")
        try:
            poll_resp = httpx.post(
                "https://api.workos.com/user_management/authenticate",
                data={
                    "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
                    "device_code": device_code,
                    "client_id": WORKOS_CLIENT_ID,
                },
                timeout=30,
            )
        except httpx.ReadTimeout:
            print("  Poll timed out, retrying...")
            continue

        print(f"  Response: {poll_resp.status_code} {poll_resp.text[:200]}")

        if poll_resp.status_code == 200:
            token_data = poll_resp.json()
            email = token_data["user"]["email"]
            workos_user_id = token_data["user"]["id"]
            break

        error_body = poll_resp.json()
        error_code = error_body.get("error", "")

        if error_code == "authorization_pending":
            continue
        elif error_code == "slow_down":
            interval += 1
            continue
        else:
            print(f"  Unexpected error: {error_code} — {poll_resp.text[:300]}")
            poll_resp.raise_for_status()

    print(f"  WorkOS auth success: {email}")
    print(f"  Calling {server_url}/auth/login ...")
    login_resp = httpx.post(
        f"{server_url}/auth/login",
        json={"email": email, "workos_user_id": workos_user_id},
        timeout=30,
    )
    print(f"  Login response: {login_resp.status_code} {login_resp.text[:200]}")
    login_resp.raise_for_status()
    api_key = login_resp.json()["api_key"]

    return (api_key, email)


def _action_manage_users(server_url: str, api_key: str) -> None:
    headers = _api_headers(api_key)

    while True:
        action = inquirer.select(
            message="User management:",
            choices=["Create user", "Delete user", "Manage user groups", "Back"],
        ).execute()

        if action == "Back":
            break

        if action == "Create user":
            display_name = inquirer.text(message="Display name:").execute()
            if not display_name:
                continue
            resp = httpx.post(
                f"{server_url}/auth/users",
                headers=headers,
                json={"display_name": display_name},
                timeout=10,
            )
            resp.raise_for_status()
            user = resp.json()
            print(f"\n  Created user: {user}\n")

        elif action == "Delete user":
            resp = httpx.get(f"{server_url}/auth/users", headers=headers, timeout=10)
            resp.raise_for_status()
            users = resp.json()
            if not users:
                print("\n  No users found.\n")
                continue
            choices = [
                {"name": u.get("display_name", u["user_id"]), "value": u["user_id"]}
                for u in users
            ]
            user_id = inquirer.select(message="Select user to delete:", choices=choices).execute()
            httpx.delete(
                f"{server_url}/auth/users/{user_id}", headers=headers, timeout=10
            ).raise_for_status()
            print(f"\n  Deleted user {user_id}.\n")

        elif action == "Manage user groups":
            resp = httpx.get(f"{server_url}/auth/users", headers=headers, timeout=10)
            resp.raise_for_status()
            users = resp.json()
            if not users:
                print("\n  No users found.\n")
                continue
            choices = [
                {"name": u.get("display_name", u["user_id"]), "value": u["user_id"]}
                for u in users
            ]
            user_id = inquirer.select(message="Select user:", choices=choices).execute()

            group_action = inquirer.select(
                message="Group action:",
                choices=["Add to group", "Remove from group", "Back"],
            ).execute()

            if group_action == "Add to group":
                remote_groups = _fetch_groups(server_url)
                group_choices = [{"name": g, "value": g} for g in remote_groups]
                group_choices.append({"name": "-- Create new group --", "value": "__create__"})

                selected = inquirer.select(
                    message="Select group:", choices=group_choices
                ).execute()

                if selected == "__create__":
                    group_name = inquirer.text(message="New group name:").execute()
                    if not group_name:
                        continue
                    password = _create_group(server_url, group_name)
                    if password is None:
                        continue
                    selected = group_name
                    print(f"\n  Created group '{group_name}' with password: {password}\n")
                else:
                    password = inquirer.secret(
                        message=f"Password for '{selected}':"
                    ).execute()

                httpx.post(
                    f"{server_url}/auth/users/{user_id}/groups",
                    headers=headers,
                    json={"group_id": selected, "password": password},
                    timeout=10,
                ).raise_for_status()
                print(f"\n  Added user to group '{selected}'.\n")

            elif group_action == "Remove from group":
                ug_resp = httpx.get(
                    f"{server_url}/auth/users/{user_id}/groups",
                    headers=headers,
                    timeout=10,
                )
                ug_resp.raise_for_status()
                user_groups = ug_resp.json().get("groups", [])
                if not user_groups:
                    print("\n  User is not in any groups.\n")
                    continue
                group_to_remove = inquirer.select(
                    message="Select group to remove from:",
                    choices=user_groups,
                ).execute()
                httpx.delete(
                    f"{server_url}/auth/users/{user_id}/groups/{group_to_remove}",
                    headers=headers,
                    timeout=10,
                ).raise_for_status()
                print(f"\n  Removed user from group '{group_to_remove}'.\n")


def _setup_tui_sidebar(
    target_pane: str,
    server_url: str,
    user_id: str,
    api_key: str,
) -> str | None:
    split_result = subprocess.run(
        [
            "tmux", "split-window", "-h",
            "-t", target_pane,
            "-l", "40%",
            "-P", "-F", "#{session_name}:#{window_index}.#{pane_index}",
        ],
        capture_output=True,
        text=True,
    )
    if split_result.returncode != 0:
        print(f"  Failed to split tmux window: {split_result.stderr}", file=sys.stderr)
        return None

    new_pane = split_result.stdout.strip()
    cmd = (
        f"relay-tui "
        f"--server-url {server_url} "
        f"--user-id {user_id} "
        f"--api-key {api_key} "
        f"--target-pane {target_pane}"
    )
    subprocess.run(["tmux", "send-keys", "-t", new_pane, cmd, "Enter"], check=True)
    subprocess.run(["tmux", "select-pane", "-t", target_pane], check=True)
    return new_pane


def _action_setup_sessions(server_url: str, api_key: str) -> None:
    headers = _api_headers(api_key)

    resp = httpx.get(f"{server_url}/auth/users", headers=headers, timeout=10)
    resp.raise_for_status()
    users = resp.json()

    if not users:
        print("\n  No users found. Create users first.\n")
        return

    sessions = _discover_claude_sessions()
    if not sessions:
        launch = inquirer.confirm(
            message="No Claude sessions found. Launch a new one?", default=True
        ).execute()
        if launch:
            new_pane = _launch_new_claude_session()
            if new_pane:
                sessions = _discover_claude_sessions()
        if not sessions:
            return

    used_users: set[str] = set()
    used_sessions: set[str] = set()
    mappings: list[tuple[dict[str, Any], dict[str, str]]] = []

    while True:
        available_users = [u for u in users if u["user_id"] not in used_users]
        available_sessions = [s for s in sessions if s["pane_id"] not in used_sessions]

        if not available_users or not available_sessions:
            break

        user_choices = [
            {"name": u.get("display_name", u["user_id"]), "value": u["user_id"]}
            for u in available_users
        ]
        user_choices.append({"name": "-- Done mapping --", "value": "__done__"})

        selected_user_id = inquirer.select(
            message="Select a user to map:", choices=user_choices
        ).execute()
        if selected_user_id == "__done__":
            break

        session_choices = [
            {"name": f"{s['pane_id']}  (pid {s['pid']})", "value": s["pane_id"]}
            for s in available_sessions
        ]
        selected_pane = inquirer.select(
            message="Select a session:", choices=session_choices
        ).execute()

        selected_user = next(u for u in users if u["user_id"] == selected_user_id)
        selected_session = next(s for s in sessions if s["pane_id"] == selected_pane)

        used_users.add(selected_user_id)
        used_sessions.add(selected_pane)
        mappings.append((selected_user, selected_session))

    for user, session in mappings:
        pane_id = session["pane_id"]
        user_id = user["user_id"]
        display_name = user.get("display_name", user_id)

        open_tui = inquirer.confirm(
            message=f"Open TUI sidebar for {display_name}?",
            default=True,
        ).execute()

        if open_tui:
            tui_pane = _setup_tui_sidebar(pane_id, server_url, user_id, api_key)
            if tui_pane:
                print(f"  TUI running in pane: {tui_pane}")

    for user, session in mappings:
        pane_id = session["pane_id"]
        user_id = user["user_id"]
        display_name = user.get("display_name", user_id)

        text = (
            f'[ALSO SETUP] You are connected to the Also relay as "{display_name}". '
            f'Your user_id is "{user_id}". Pass this user_id as the first argument '
            f'to every relay tool call (send_message, read_new_messages, etc).'
        )
        subprocess.run(["tmux", "send-keys", "-t", pane_id, "-l", text])
        subprocess.run(["tmux", "send-keys", "-t", pane_id, "Enter"])
        print(f"  Injected setup message into {pane_id} for {display_name}")


def _action_show_status(server_url: str, api_key: str) -> None:
    masked_key = api_key[:8] + "..." if len(api_key) > 8 else api_key
    print(f"\n  Server:   {server_url}")
    print(f"  API Key:  {masked_key}")
    headers = _api_headers(api_key)
    try:
        resp = httpx.get(f"{server_url}/auth/users", headers=headers, timeout=10)
        resp.raise_for_status()
        users = resp.json()
        if users:
            print("  Users:")
            for u in users:
                groups_resp = httpx.get(
                    f"{server_url}/auth/users/{u['user_id']}/groups",
                    headers=headers,
                    timeout=10,
                )
                groups = groups_resp.json().get("groups", []) if groups_resp.status_code == 200 else []
                groups_str = ", ".join(groups) if groups else "(no groups)"
                print(f"    {u.get('display_name', u['user_id'])} [{u['user_id']}] -> {groups_str}")
        else:
            print("  Users:    (none)")
    except httpx.HTTPError:
        print("  Users:    (could not fetch)")
    print()


def main() -> None:
    if len(sys.argv) > 1 and sys.argv[1] in ("--version", "-v"):
        print(f"alsogc {_get_version()}")
        return

    api_key = _get_api_key()
    server_url = _get_server_url()

    if not api_key:
        print("  No API key found. Authenticating...")
        api_key, email = _authenticate_workos(server_url)
        _sync_mcp_config(server_url, api_key)
        print(f"  Authenticated as {email}.\n")
    else:
        _sync_mcp_config(server_url, api_key)
        print(f"  Synced also-mcp server in {CLAUDE_CONFIG_PATH}\n")

    while True:
        action = inquirer.select(
            message="What would you like to do?",
            choices=[
                "Manage users",
                "Setup sessions",
                "Status",
                "Exit",
            ],
        ).execute()

        if action == "Manage users":
            _action_manage_users(server_url, api_key)
        elif action == "Setup sessions":
            _action_setup_sessions(server_url, api_key)
        elif action == "Status":
            _action_show_status(server_url, api_key)
        else:
            break


if __name__ == "__main__":
    main()



from datetime import UTC, datetime

import pytest

from relay.server.models import Group, Message, Peer


class TestPeer:
    def test_requires_fields(self):
        peer = Peer(user_id="u-1", display_name="Alice", group_id="g1")
        assert peer.user_id == "u-1"
        assert peer.display_name == "Alice"
        assert peer.group_id == "g1"

    def test_defaults(self):
        peer = Peer(user_id="u-1", display_name="Alice", group_id="g1")
        assert peer.last_read_index == 0
        assert peer.online is False

    def test_explicit_fields(self):
        peer = Peer(
            user_id="u-42",
            display_name="Bob",
            group_id="team1",
            last_read_index=10,
            online=True,
        )
        assert peer.user_id == "u-42"
        assert peer.display_name == "Bob"
        assert peer.group_id == "team1"
        assert peer.last_read_index == 10
        assert peer.online is True

    def test_model_is_mutable(self):
        peer = Peer(user_id="u-1", display_name="Alice", group_id="g1")
        peer.online = True
        assert peer.online is True
        peer.last_read_index = 5
        assert peer.last_read_index == 5


class TestMessage:
    def test_auto_generates_message_id(self):
        msg = Message(
            from_peer="p1", from_display="host:proj", content="hello", group_id="g1"
        )
        assert len(msg.message_id) == 32
        assert msg.message_id.isalnum()

    def test_auto_generates_timestamp(self):
        before = datetime.now(UTC)
        msg = Message(
            from_peer="p1", from_display="host:proj", content="hello", group_id="g1"
        )
        after = datetime.now(UTC)
        assert before <= msg.timestamp <= after

    def test_model_is_frozen(self):
        msg = Message(
            from_peer="p1", from_display="host:proj", content="hello", group_id="g1"
        )
        with pytest.raises(Exception):
            msg.content = "changed"

    def test_unique_message_ids(self):
        msgs = [
            Message(from_peer="p1", from_display="d", content="hi", group_id="g1")
            for _ in range(50)
        ]
        ids = {m.message_id for m in msgs}
        assert len(ids) == 50


class TestGroup:
    def test_empty_group(self):
        group = Group(group_id="g1")
        assert group.group_id == "g1"
        assert group.peers == {}
        assert group.messages == []

    def test_add_peer_to_group(self):
        group = Group(group_id="g1")
        peer = Peer(user_id="u-1", display_name="Alice", group_id="g1")
        group.peers[peer.user_id] = peer
        assert peer.user_id in group.peers

    def test_add_message_to_group(self):
        group = Group(group_id="g1")
        msg = Message(
            from_peer="p1", from_display="host:proj", content="hello", group_id="g1"
        )
        group.messages.append(msg)
        assert len(group.messages) == 1
        assert group.messages[0].content == "hello"



from __future__ import annotations

import asyncio
import json
from collections.abc import AsyncIterator
from typing import TYPE_CHECKING, Any

from fastmcp import FastMCP
from sse_starlette.sse import EventSourceResponse, ServerSentEvent
from starlette.requests import Request
from starlette.responses import JSONResponse, Response

from relay.server.state import RelayState

if TYPE_CHECKING:
    from relay.server.db import RelayDB


async def notification_stream(
    state: RelayState, user_id: str
) -> AsyncIterator[ServerSentEvent]:
    queue = await state.connect_sse(user_id)
    try:
        while True:
            payload: dict[str, Any] = await queue.get()
            yield ServerSentEvent(
                data=json.dumps(payload),
                event="notification",
                id=payload["message_id"],
            )
    except asyncio.CancelledError:
        pass
    finally:
        await state.disconnect_sse(user_id)


def register_sse_routes(
    mcp: FastMCP, state: RelayState, db: RelayDB | None = None
) -> None:
    @mcp.custom_route("/notifications/{user_id}", methods=["GET"])
    async def notifications_endpoint(request: Request) -> Response:
        user_id = request.path_params["user_id"]
        if db is not None:
            api_key = request.headers.get("x-api-key", "")
            if not api_key:
                return JSONResponse({"error": "unauthorized"}, status_code=401)
            account = await db.get_account_by_api_key(api_key)
            if account is None:
                return JSONResponse({"error": "unauthorized"}, status_code=401)
            user = await db.get_user(user_id)
            if user is None or user.account_email != account.email:
                return JSONResponse({"error": "forbidden"}, status_code=403)
        return EventSourceResponse(notification_stream(state, user_id))



from __future__ import annotations

import json
import os
import secrets
import time
from collections import OrderedDict
from typing import Any, Generic, TypeVar
from uuid import uuid4

import bcrypt
import httpx
from fastmcp import FastMCP
from starlette.requests import Request
from starlette.responses import JSONResponse, Response
from starlette.types import ASGIApp, Receive, Scope, Send

from relay.server.db import AccountRow, RelayDB

K = TypeVar("K")
V = TypeVar("V")


class BoundedTTLCache(Generic[K, V]):
    def __init__(self, maxsize: int, ttl: float) -> None:
        self._maxsize = maxsize
        self._ttl = ttl
        self._data: OrderedDict[K, tuple[V, float]] = OrderedDict()

    def get(self, key: K) -> V | None:
        entry = self._data.get(key)
        if entry is None:
            return None
        value, ts = entry
        if time.monotonic() - ts > self._ttl:
            del self._data[key]
            return None
        self._data.move_to_end(key)
        return value

    def put(self, key: K, value: V) -> None:
        if key in self._data:
            self._data.move_to_end(key)
            self._data[key] = (value, time.monotonic())
            return
        if len(self._data) >= self._maxsize:
            self._data.popitem(last=False)
        self._data[key] = (value, time.monotonic())


PROTECTED_PREFIXES = ("/mcp", "/notifications")
PUBLIC_PREFIXES = ("/auth",)

_CACHE_MAX_SIZE = 1024
_CACHE_TTL_SECONDS = 300.0


class ApiKeyAuthMiddleware:
    def __init__(self, app: ASGIApp, db: RelayDB) -> None:
        self._app = app
        self._db = db
        self._key_cache: BoundedTTLCache[str, str] = BoundedTTLCache(
            maxsize=_CACHE_MAX_SIZE, ttl=_CACHE_TTL_SECONDS
        )

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] != "http":
            await self._app(scope, receive, send)
            return

        path: str = scope.get("path", "")

        if any(path.startswith(p) for p in PUBLIC_PREFIXES):
            await self._app(scope, receive, send)
            return

        if not any(path.startswith(p) for p in PROTECTED_PREFIXES):
            await self._app(scope, receive, send)
            return

        headers = dict(scope.get("headers", []))
        api_key = headers.get(b"x-api-key", b"").decode()

        if not api_key:
            response = JSONResponse({"error": "unauthorized"}, status_code=401)
            await response(scope, receive, send)
            return

        email = self._key_cache.get(api_key)
        if email is None:
            account = await self._db.get_account_by_api_key(api_key)
            if account is None:
                response = JSONResponse({"error": "unauthorized"}, status_code=401)
                await response(scope, receive, send)
                return
            email = account.email
            self._key_cache.put(api_key, email)

        scope["state"] = scope.get("state", {})
        scope["state"]["account_email"] = email

        await self._app(scope, receive, send)


async def _authenticate_api_key(request: Request, db: RelayDB) -> AccountRow | None:
    api_key = request.headers.get("x-api-key", "")
    if not api_key:
        return None
    return await db.get_account_by_api_key(api_key)


def register_auth_routes(mcp: FastMCP, db: RelayDB) -> None:
    @mcp.custom_route("/auth/groups", methods=["GET"])
    async def list_groups(request: Request) -> Response:
        groups = await db.list_groups()
        return JSONResponse(groups)

    @mcp.custom_route("/auth/groups", methods=["POST"])
    async def create_group(request: Request) -> Response:
        try:
            body: dict[str, Any] = json.loads(await request.body())
        except json.JSONDecodeError:
            return JSONResponse({"error": "invalid json"}, status_code=400)
        group_id = body.get("group_id", "")
        if not group_id:
            return JSONResponse(
                {"error": "group_id is required"}, status_code=400
            )

        if await db.group_exists(group_id):
            return JSONResponse(
                {"error": "group already exists"}, status_code=409
            )

        password = secrets.token_urlsafe(6)
        password_hash = bcrypt.hashpw(
            password.encode("utf-8"), bcrypt.gensalt()
        ).decode("utf-8")

        await db.create_group(group_id, password_hash)
        return JSONResponse({"group_id": group_id, "password": password})

    @mcp.custom_route("/auth/groups/{group_id}/verify", methods=["POST"])
    async def verify_group(request: Request) -> Response:
        group_id = request.path_params["group_id"]
        try:
            body: dict[str, Any] = json.loads(await request.body())
        except json.JSONDecodeError:
            return JSONResponse({"error": "invalid json"}, status_code=400)
        password = body.get("password", "")

        stored_hash = await db.get_group_password_hash(group_id)
        if stored_hash is None:
            return JSONResponse(
                {"ok": False, "error": "group not found"}, status_code=404
            )

        ok = bcrypt.checkpw(
            password.encode("utf-8"), stored_hash.encode("utf-8")
        )
        return JSONResponse({"ok": ok})

    @mcp.custom_route("/auth/login", methods=["POST"])
    async def login(request: Request) -> Response:
        try:
            body: dict[str, Any] = json.loads(await request.body())
        except json.JSONDecodeError:
            return JSONResponse({"error": "invalid json"}, status_code=400)

        email = body.get("email", "")
        workos_user_id = body.get("workos_user_id", "")
        if not email or not workos_user_id:
            return JSONResponse({"error": "email and workos_user_id required"}, status_code=400)

        workos_api_key = os.environ.get("WORKOS_API_KEY", "")
        if not workos_api_key:
            return JSONResponse({"error": "server misconfigured"}, status_code=500)

        async with httpx.AsyncClient() as client:
            resp = await client.get(
                f"https://api.workos.com/user_management/users/{workos_user_id}",
                headers={"Authorization": f"Bearer {workos_api_key}"},
            )

        if resp.status_code != 200:
            return JSONResponse({"error": "invalid workos user"}, status_code=401)

        workos_data = resp.json()
        workos_email = workos_data.get("email", "")
        if workos_email != email:
            return JSONResponse({"error": "email mismatch"}, status_code=401)

        account = await db.get_account_by_email(email)
        if account is not None:
            return JSONResponse({"api_key": account.api_key, "email": email})

        api_key = secrets.token_urlsafe(32)
        await db.create_account(email, api_key)
        return JSONResponse({"api_key": api_key, "email": email})

    @mcp.custom_route("/auth/users", methods=["GET"])
    async def list_users(request: Request) -> Response:
        account = await _authenticate_api_key(request, db)
        if account is None:
            return JSONResponse({"error": "unauthorized"}, status_code=401)

        users = await db.list_users(account.email)
        return JSONResponse([
            {"user_id": u.user_id, "display_name": u.display_name}
            for u in users
        ])

    @mcp.custom_route("/auth/users", methods=["POST"])
    async def create_user(request: Request) -> Response:
        account = await _authenticate_api_key(request, db)
        if account is None:
            return JSONResponse({"error": "unauthorized"}, status_code=401)

        try:
            body: dict[str, Any] = json.loads(await request.body())
        except json.JSONDecodeError:
            return JSONResponse({"error": "invalid json"}, status_code=400)

        display_name = body.get("display_name", "")
        if not display_name:
            return JSONResponse({"error": "display_name is required"}, status_code=400)

        user_id = f"u-{uuid4().hex[:8]}"
        await db.create_user(user_id, account.email, display_name)
        return JSONResponse({"user_id": user_id, "display_name": display_name})

    @mcp.custom_route("/auth/users/{user_id}", methods=["DELETE"])
    async def delete_user(request: Request) -> Response:
        account = await _authenticate_api_key(request, db)
        if account is None:
            return JSONResponse({"error": "unauthorized"}, status_code=401)

        user_id = request.path_params["user_id"]
        user = await db.get_user(user_id)
        if user is None or user.account_email != account.email:
            return JSONResponse({"error": "not found"}, status_code=404)

        await db.delete_user(user_id)
        return JSONResponse({"ok": True})

    @mcp.custom_route("/auth/users/{user_id}/groups", methods=["GET"])
    async def get_user_groups(request: Request) -> Response:
        account = await _authenticate_api_key(request, db)
        if account is None:
            return JSONResponse({"error": "unauthorized"}, status_code=401)

        user_id = request.path_params["user_id"]
        user = await db.get_user(user_id)
        if user is None or user.account_email != account.email:
            return JSONResponse({"error": "not found"}, status_code=404)

        groups = await db.get_user_groups(user_id)
        return JSONResponse({"groups": groups})

    @mcp.custom_route("/auth/users/{user_id}/groups", methods=["POST"])
    async def add_user_to_group(request: Request) -> Response:
        account = await _authenticate_api_key(request, db)
        if account is None:
            return JSONResponse({"error": "unauthorized"}, status_code=401)

        user_id = request.path_params["user_id"]
        user = await db.get_user(user_id)
        if user is None or user.account_email != account.email:
            return JSONResponse({"error": "not found"}, status_code=404)

        try:
            body: dict[str, Any] = json.loads(await request.body())
        except json.JSONDecodeError:
            return JSONResponse({"error": "invalid json"}, status_code=400)

        group_id = body.get("group_id", "")
        password = body.get("password", "")

        stored_hash = await db.get_group_password_hash(group_id)
        if stored_hash is None:
            return JSONResponse({"error": "group not found"}, status_code=404)

        if not bcrypt.checkpw(password.encode("utf-8"), stored_hash.encode("utf-8")):
            return JSONResponse({"error": "invalid password"}, status_code=401)

        await db.add_user_to_group(user_id, group_id)
        return JSONResponse({"ok": True})

    @mcp.custom_route("/auth/users/{user_id}/groups/{group_id}", methods=["DELETE"])
    async def remove_user_from_group(request: Request) -> Response:
        account = await _authenticate_api_key(request, db)
        if account is None:
            return JSONResponse({"error": "unauthorized"}, status_code=401)

        user_id = request.path_params["user_id"]
        user = await db.get_user(user_id)
        if user is None or user.account_email != account.email:
            return JSONResponse({"error": "not found"}, status_code=404)

        group_id = request.path_params["group_id"]
        await db.remove_user_from_group(user_id, group_id)
        return JSONResponse({"ok": True})
